#include <process.h>
#include <windows.h>
#include <vector>
#include <string>
#include <sstream>
#include <angelscript.h>

#include "application.h"
#include "imagewnd.h"
#include "outstream.h"
#include "sysfuncs.h"
#include "file.h"
using namespace std;

CApplication app;

CApplication::CApplication()
{
	isExecuting = false;
	isWorking = false;
	scriptEngine = 0;
}

int CApplication::Run()
{
	scriptEngine = asCreateScriptEngine(ANGELSCRIPT_VERSION);
	if( scriptEngine == 0 )
	{
		MessageBox(0, "Failed to create script engine.\n", "Error", MB_OK);
		return -1;
	}

	if( wnd.Create() < 0 )
	{
		MessageBox(0, "Failed to create window\n", "Error", MB_OK);
		return -1;
	}

	RegisterSystemFunctions(scriptEngine);

	while( !wnd.CheckMessages(true) );

	FreeMemory();

	if( scriptEngine )
	{
		scriptEngine->Release();
		scriptEngine = 0;
	}

	return 0;
}

void CApplication::SetScript(const char *file)
{
	scriptFile = file;

	COutStream out(&wnd.edit);
	out.Write(("Script file: " + scriptFile + "\n\n").c_str());

	scriptOK = false;
}

int CApplication::CompileScript()
{
	if( isWorking ) return -1;

	if( scriptFile == "" )
	{
		MessageBox(wnd.GetHandle(), "Please choose a script first.", "Info", MB_OK|MB_ICONINFORMATION);
		return 0;
	}

	isWorking = true;

	_beginthread(CompileThread, 0, 0);

	return 0;
}

int CApplication::ExecuteScript()
{
	if( isWorking ) return -1;

	if( scriptFile == "" || !scriptOK )
	{
		MessageBox(wnd.GetHandle(), "A script must be compiled without errors first.", "Info", MB_OK|MB_ICONINFORMATION);
		return 0;
	}

	isWorking = true;

	FreeMemory();

	_beginthread(ExecuteThread, 0, 0);

	return 0;
}

void CApplication::AbortWork()
{
	doAbortWork = true;

	app.context->Abort();
}

void CApplication::LineCallback(asIScriptContext *ctx)
{
	int column;
	int line = ctx->GetCurrentLineNumber(&column);

	string func = scriptEngine->GetFunctionDeclaration(ctx->GetCurrentFunction());

	int callstackSize = ctx->GetCallstackSize();
}

void CApplication::ExceptionCallback(asIScriptContext *ctx)
{
	asIScriptEngine *engine = ctx->GetEngine();
	int funcID = ctx->GetExceptionFunction();

	COutStream out(&wnd.edit);
	stringstream str;

	out.Write("--- exception ---\n");
	str.seekp(0); str << "desc: " << ctx->GetExceptionString() << endl << ends;
	out.Write(str.str().c_str());
	str.seekp(0); str << "func: " << engine->GetFunctionDeclaration(funcID) << endl << ends;
	out.Write(str.str().c_str());
	str.seekp(0); str << "modl: " << engine->GetModuleNameFromIndex(asMODULEIDX(funcID)) << endl << ends;
	out.Write(str.str().c_str());
	str.seekp(0); str << "sect: " << engine->GetFunctionSection(funcID) << endl << ends;
	out.Write(str.str().c_str());
	int col, line = ctx->GetExceptionLineNumber(&col);
	str.seekp(0); str << "line: " << line << ", " << col << endl << ends;
	out.Write(str.str().c_str());

	out.Write("--- callstack ---\n");
	for( int n = 0; n < ctx->GetCallstackSize(); n++ )
	{
		funcID = ctx->GetCallstackFunction(n);
		line = ctx->GetCallstackLineNumber(n,&col);
		str.seekp(0); str << engine->GetModuleNameFromIndex(asMODULEIDX(funcID)) << ":" << engine->GetFunctionDeclaration(funcID) << ":" << line << ", " << col << endl << ends;
		out.Write(str.str().c_str());
	}
	out.Write("-----------------\n");
}

void ExecuteThread(void *data)
{
	app.isWorking = true;
	app.isExecuting = true;
	app.doAbortWork = false;

	COutStream out(&app.wnd.edit);

	DWORD milli = timeGetTime();

	out.Write("Executing...\n");

	app.scriptEngine->CreateContext(&app.context);
	if( (app.context->Prepare(app.scriptEngine->GetFunctionIDByName(0, "main"))) >= 0 )
	{
		// Uncomment this line to be able to follow the script execution statement by statement
//		app.context->SetLineCallback(asMETHOD(CApplication,LineCallback), &app, asCALL_THISCALL);
		app.context->SetExceptionCallback(asMETHOD(CApplication,ExceptionCallback), &app, asCALL_THISCALL);

		int r = app.context->Execute();
		if( r == 3 )
			out.Write("Ended with exception.\n");
		else if( r == 2 )
		{
			if( app.doAbortWork )
				out.Write("Aborted as requested.\n");
			else
				out.Write("Aborted without user request.\n");
		}
		else if( r == 0 )
			out.Write("Ended succesfully.\n");
		else if( r == 1 )
			out.Write("Program suspended.\n");
		else
			out.Write("Execution failed.\n");
	}
	app.context->Release();

	milli = timeGetTime() - milli;
	stringstream str;
	str << "time: " << milli << " ms" << endl << endl;
	out.Write(str.str().c_str());

	app.isExecuting = false;
	app.isWorking = false;
}

void CompileThread(void *data)
{
	app.isWorking = true;

	app.scriptOK = false;

	DWORD milli = timeGetTime();

	COutStream out(&app.wnd.edit);

	out.Write("Building...\n");

	// Load the script code
	vector<char> script;
	CFile file;
	if( file.Open(app.scriptFile.c_str(), "rb") < 0 )
	{
		out.Write("Failed to open script file.\n");
		app.isWorking = false;
		return;
	}

	script.resize(file.GetSize());

	file.Read(&script[0], 1, script.size());
	file.Close();

	app.scriptEngine->AddScriptSection(0, app.scriptFile.c_str(), &script[0], script.size());
	int r = app.scriptEngine->Build(0, &out);

	if( r < 0 )
		out.Write("Build failed.\n");
	else
		out.Write("Ended succesfully.\n");

	milli = timeGetTime() - milli;
	stringstream str;
	str << "time: " << milli << " ms" << endl << endl;

	out.Write(str.str().c_str());

	// Check interface for main function
	if( r >= 0 )
	{
		string decl = app.scriptEngine->GetFunctionDeclaration(app.scriptEngine->GetFunctionIDByName(0, "main"));

		if( decl != "void main()" )
			MessageBox(0, "Declaration of main must be \"void main()\".", "Script error", MB_OK);
		else
			app.scriptOK = true;
	}
	else
		MessageBox(0, "The script has errors. Correct them and try again.", "Script error", MB_OK);
	
	app.isWorking = false;
}

void CApplication::ShowImage(CFloatImage *img)
{
	if( img == 0 ) return;

	CImageWnd *imageWnd;

	imageWnd = new CImageWnd;

	// Copy image to the window
	imageWnd->CopyImage(img);

	imageWindows.push_back(imageWnd);

	PostMessage(wnd.GetHandle(), MSG_SHOWIMAGE, (WPARAM)(imageWindows.size()-1), 0);
}

void CApplication::OpenImageWindow(int imageWindow)
{
	if( imageWindows[imageWindow] )
	{
		imageWindows[imageWindow]->Create(&wnd);
	}
}

void CApplication::FreeMemory()
{
	// Delete image windows
	for( int n = 0; n < imageWindows.size(); n++ )
	{
		if( imageWindows[n] )
		{
			delete imageWindows[n];
			imageWindows[n] = 0;
		}
	}

	imageWindows.clear();
}
