#include <io.h> // _filelength()
#include <stdarg.h>		// va_list, va_start(), etc
#include "file.h"


//=============================================================================

CFile::CFile()
{
	file = 0;
}

//=============================================================================

CFile::~CFile()
{
	Close();
}

//=============================================================================

// See fopen() for description
int CFile::Open(const char *fileName, const char *mode)
{
	if( file != 0 )
		return EFILE_INVALID_CALL;

	int lowFlags = 0;

	file = fopen(fileName, mode);
	if( file == 0 )
		return EFILE_FAILED;

	return 0;
}

//=============================================================================

int CFile::Close()
{
	if( file != 0 )
	{
		int r = fclose(file);

		file = 0;

		if( r == EOF )
			return EFILE_FAILED;
	}

	return 0;
}

//=============================================================================

bool CFile::IsEOF()
{
	if( file != 0 )
	{
		int r = feof(file);
		if( r == 0 )
			return false;
	}

	return true;
}

//=============================================================================

// Returns number of complete items read, or EFILE_INVALID_CALL
int CFile::Read(void *data, int size, int count)
{
	if( file == 0 )
		return EFILE_INVALID_CALL;
	
	return fread(data, size, count, file);
}

//=============================================================================

// Returns the byte read or EOF (-1)
int CFile::ReadByte()
{
	if( file == 0 ) return EOF;

	return fgetc(file);
}

//=============================================================================

// Returns number of items written or EFILE_INVALID_CALL
int CFile::Write(const void *data, int size, int count)
{
	if( file == 0 )
		return EFILE_INVALID_CALL;

	return fwrite(data, size, count, file);
}

int CFile::WriteByte(BYTE byte)
{
	if( file == 0 )
		return EFILE_INVALID_CALL;

	int r = fputc(byte, file);
	if( r == EOF )
		return EFILE_FAILED;

	return 0;
}

int CFile::FormatWrite(const char *fmt, ...)
{
	if( file == 0 )
		return EFILE_INVALID_CALL;

	va_list args;
	va_start(args, fmt);

	int r = vfprintf(file, fmt, args);

	va_end(args);

	if( r < 0 )
		return EFILE_FAILED;

	return r;
}

//=============================================================================

// Returns number of bytes read or an error code. The delimiter is also 
// returned in the buffer.
int CFile::ReadUntil(void *data, long maxBytes, BYTE delimiter)
{
	if( file == 0 )
		return EFILE_INVALID_CALL;

	// Read bytes until delimiter has been found
	for( long n = 0; n < maxBytes; n++ )
	{
		int r = fgetc(file);
		if( r == EOF ) break;

		((BYTE*)data)[n] = BYTE(r);

		if( r == delimiter ) return ++n;
	}

	return n;
}

//=============================================================================

// Returns number of bytes skipped or an error code. 
// The delimiter is also skipped.
int CFile::SkipUntil(BYTE delimiter)
{
	if( file == 0 )
		return EFILE_INVALID_CALL;

	// Read bytes until delimiter has been found
	for( long n = 0; ; n++ )
	{
		int r = ::fgetc(file);
		if( r == EOF ) break;
		if( r == delimiter ) return ++n;
	}

	return n;
}

// Skips all bytes in theseBytes and returns the number of bytes skipped
int CFile::Skip(const BYTE *theseBytes, int numBytes)
{
	if( file == 0 )
		return EFILE_INVALID_CALL;

	// Read bytes until it is no longer one of theseBytes
	for( long n = 0; ; n++ )
	{
		int r = ::fgetc(file);
		if( r == EOF ) break;
		for( int c = 0; c < numBytes; c++ )
			if( r == theseBytes[c] ) break;
		if( c == numBytes ) 
		{
			// Return this character to the stream
			::fseek(file, -1, SEEK_CUR);
			break;
		}
	}

	return n;
}

//=============================================================================

int CFile::Seek(long offset, int origin)
{
	if( file == 0 )
		return EFILE_INVALID_CALL;

	int r = ::fseek(file, offset, origin);
	if( r < 0 )
		return EFILE_FAILED;

	return 0;
}

int CFile::Tell()
{
	if( file == 0 )
		return EFILE_INVALID_CALL;

	int r = ::ftell(file);
	if( r < 0 )
		return EFILE_FAILED;

	return r;
}

int CFile::GetSize()
{
	if( file == 0 )
		return EFILE_INVALID_CALL;

	long len = ::_filelength( _fileno(file) );
	
	if( len == -1 ) return EFILE_FAILED;

	return len;
}